#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#define BYTE uint8_t
#define WORD uint16_t
#define DWORD uint32_t

#define CHAN_NONE 17

/* ------------------------------------------------------------------------- */

static void *lib_malloc(unsigned int size)
{
    void *ptr = malloc(size);

    if (ptr == NULL) {
        exit(EXIT_FAILURE);
    }

    return ptr;
}

static void lib_free(void *ptr)
{
    free(ptr);
}

/* ------------------------------------------------------------------------- */

static WORD parse_be_w(BYTE **ptr)
{
    WORD v = ((*ptr)[0] << 8)
           | (*ptr)[1];
    *ptr += 2;
    return v;
}

static DWORD parse_be_dw(BYTE **ptr)
{
    DWORD v = ((*ptr)[0] << 24)
            | ((*ptr)[1] << 16)
            | ((*ptr)[2] << 8)
            | (*ptr)[3];
    *ptr += 4;
    return v;
}

/* ------------------------------------------------------------------------- */

static long parse_midi_header(BYTE *midi_data, long midi_data_len, WORD *ticks_out)
{
    BYTE *p = midi_data;
    DWORD hlen;
    WORD type;
    WORD tracks;
    WORD ticks;
    DWORD tlen;
    long i;

    if (memcmp(p, "MThd", 4) != 0) {
        fprintf(stderr, "invalid header\n");
        return -1;
    }
    p += 4;

    hlen = parse_be_dw(&p);
    type = parse_be_w(&p);
    tracks = parse_be_w(&p);
    ticks = parse_be_w(&p);

    if (memcmp(p, "MTrk", 4) != 0) {
        fprintf(stderr, "invalid track header\n");
        return -1;
    }
    p += 4;

    tlen = parse_be_dw(&p);

    fprintf(stderr, "hlen %u, type %u, tracks %u, ticks %u, tlen %u (%li)\n", hlen, type, tracks, ticks, tlen, midi_data_len);

    if ((hlen != 6) || (type != 0) || (tracks != 1)) {
        fprintf(stderr, "unsupported\n");
        return -1;
    }

    i = p - midi_data;

    if ((long)(tlen + i) != midi_data_len) {
        fprintf(stderr, "len error %li != %li\n", (long)(tlen + i), midi_data_len);
        return -1;
    }

    *ticks_out = ticks;
    return i;
}

#define EVENT_TBL_SIZE 17
#define DELAY_TBL_SIZE 256

#define DELAY_ERROR 0xffffffff
#define MSG_ERROR 0xff
#define MSG_IGNORE 0x00

#define MSG_NOTE_ON 0x90
#define MSG_NOTE_OFF 0x80

typedef struct midi_s {
    BYTE onoff;
    BYTE chan;
    BYTE note;
    BYTE vel;
    DWORD delay;
    struct midi_s *prev;
    struct midi_s *next;
} midi_t;

static midi_t *alloc_midi(void)
{
    return lib_malloc(sizeof(midi_t));
}

static midi_t *new_midi(BYTE cmd, BYTE note, BYTE vel, DWORD delay, midi_t *prev)
{
    midi_t *m = alloc_midi();

    m->onoff = (cmd & 0x10) ? 1 : 0;
    m->chan = (cmd & 0x0f);
    m->note = note;
    m->vel = vel;
    m->delay = delay;
    m->prev = prev;
    m->next = NULL;

    if (prev != NULL) {
        m->prev->next = m;
    }
    return m;
}

static midi_t *new_midi_end(DWORD delay, midi_t *prev)
{
    midi_t *m = new_midi(0, 0, 0, delay, prev);
    m->onoff = 2/*end*/;
    return m;
}

static void free_midi_list(midi_t *midi)
{
    midi_t *temp = NULL;

    while (midi != NULL) {
        temp = midi;
        midi = midi->next;
        lib_free(temp);
    }
}

static void free_midi_list_reverse(midi_t *midi)
{
    midi_t *temp = NULL;

    while (midi != NULL) {
        temp = midi;
        midi = midi->prev;
        lib_free(temp);
    }
}

static DWORD parse_midi_delay(BYTE **ptr)
{
    BYTE *p = *ptr;
    int i = 0;
    DWORD v = 0;
    BYTE b = 0x80;

    while ((b & 0x80) == 0x80) {
        if (i >= 4) {
            fprintf(stderr, "delay too many bytes\n");
            return DELAY_ERROR;
        }
        b = p[i++];
        v <<= 7;
        v |= (b & 0x7f);
    }

    *ptr += i;
    return v;
}

static BYTE parse_midi_msg(BYTE **ptr, BYTE *note_out, BYTE *vel_out)
{
    BYTE *p = *ptr;
    int i = 0;
    BYTE b = 0x80;
    BYTE cmd = MSG_IGNORE;

    b = p[i++];
    cmd = b;

    if ((b >= 0x80) && (b < 0xf0)) {
        b &= 0xf0;
    }

    switch (b) {
        case MSG_NOTE_OFF:
        case MSG_NOTE_ON:
            *note_out = p[i++];
            *vel_out = p[i++];
            break;

        case 0xc0:
            /* 1 data, skip */
            i += 1;
            cmd = MSG_IGNORE;
            break;

        case 0xb0:
        case 0xe0:
            /* 2 data, skip */
            i += 2;
            cmd = MSG_IGNORE;
            break;

        case 0xff:
            /* skip 1 byte, get len of msg */
            b = p[++i];
            /* skip msg */
            i += b + 1;
            cmd = MSG_IGNORE;
            break;

        default:
            fprintf(stderr, "unhandled command %02X\n", b);
            return MSG_ERROR;
    }

    *ptr += i;
    return cmd;
}

static midi_t *parse_midi_events(BYTE *midi_data, long midi_data_len, long midi_data_offset, int transpose)
{
    BYTE *p = &midi_data[midi_data_offset];
    BYTE *p_end = midi_data + midi_data_len;
    midi_t *cur = NULL;
    DWORD d_acc = 0;

    while (p < p_end) {
        DWORD d;
        BYTE msg, note, vel;

        if ((d = parse_midi_delay(&p)) == DELAY_ERROR) {
            goto fail;
        }

        if ((msg = parse_midi_msg(&p, &note, &vel)) == MSG_ERROR) {
            fprintf(stderr, "...at 0x%x\n", (unsigned int)(p - midi_data));
            goto fail;
        }

        if ((d_acc + d) < d) {
            fprintf(stderr, "warn: d_acc overflow\n");
        }

        d_acc += d;

        if (msg == MSG_IGNORE) {
            continue;
        }

        cur = new_midi(msg, note + transpose, vel, d_acc, cur);
        d_acc = 0;
    }

    /* HACK add dummy tail entry */
    cur = new_midi_end(d_acc, cur);

    while ((cur != NULL) && (cur->prev != NULL)) {
        cur = cur->prev;
    }

    return cur;

fail:
    free_midi_list_reverse(cur);
    cur = NULL;
    return NULL;
}


static int parse_midi_scan_events(midi_t *midi, DWORD *have_events, DWORD *delays, DWORD divider)
{
    int num_delays = 0;
    int i;

    memset(have_events, 0, sizeof(DWORD) * EVENT_TBL_SIZE);
    memset(delays, 0xff, sizeof(DWORD) * DELAY_TBL_SIZE);

    while (midi != NULL) {
        DWORD d = midi->delay;

        ++have_events[midi->chan];

        for (i = 0; (i < num_delays) && (delays[i] < d); ++i) {
        }

        if (i == (DELAY_TBL_SIZE - 1)) {
            fprintf(stderr, "too many delays\n");
            return -1;
        }

        if (delays[i] > d) {
            memmove(&delays[i + 1], &delays[i], DELAY_TBL_SIZE - i - 1);
            delays[i] = d;
            num_delays++;
        }

        midi = midi->next;
    }

    fprintf(stderr, "channels: ");

    for (i = 0; i < 16; ++i) {
        if (have_events[i]) {
            fprintf(stderr, "%i:%i ", i, have_events[i]);
        }
    }
    fprintf(stderr, "\n");

    for (i = 0; i < num_delays; ++i) {
        fprintf(stderr, "delay %3i: %i / %i = %i rem %i\n", i, delays[i], divider, delays[i] / divider, delays[i] % divider);
    }

    return 0;
}

/* ------------------------------------------------------------------------- */

#define VPHI2PAL    (4433618.0/4.0)
#define VPHI2NTSC   (14318181.0/14.0)
#define VCH1FP (VPHI2PAL/256.0)
#define VCH2FP (VPHI2PAL/128.0)
#define VCH3FP (VPHI2PAL/64.0)
#define VCH1FN (VPHI2NTSC/256.0)
#define VCH2FN (VPHI2NTSC/128.0)
#define VCH3FN (VPHI2NTSC/64.0)

static const double vchscale[2/*pal/ntsc*/][3/*ch*/] = {
    { VCH1FP, VCH2FP, VCH3FP },
    { VCH1FN, VCH2FN, VCH3FN }
};

static double midi_freq[0x80];

static BYTE midi_to_vic_tbl[3][0x80];

static void midi_to_vic_precalc(int tv)
{
    int c, n, vv;

    for (n = 0; n < 0x80; ++n) {
        midi_freq[n] = 440.0 * powf(2.0, (n - 69) / 12.0);
    }

    for (c = 0; c < 3; ++c) {
        double vs = vchscale[tv][c];
        for (n = 0; n < 0x80; ++n) {
            double vf, mf = midi_freq[n];
            vv = 255 - round(vs / mf);
            if ((vv > 0xff) || (vv < 0x80)) {
                if (vv > 0xff) {
                    vv = 2; /* f too high for ch */
                } else {
                    vv = 1; /* f too low for ch */
                }
            }
            if (vv != 255) {
                vf = vs / (255.0 - (double)vv);
            } else {
                vf = 0.0;
            }
//            fprintf(stdout, "%i %02x -> %02x : %f\t%f\n", c, n, vv, mf, vf);
            midi_to_vic_tbl[c][n] = (BYTE)vv;
        }
    }
}

/* stores the corresponding MIDI note | 0x80 */
static BYTE note_ch_use_tbl[3][0x100];

static int output_vic_eor(int tv)
{
    int bytes = 0, c;

    for (c = 0; c < 3; ++c) {
        BYTE n, vv, vvo;
        double vs = vchscale[tv][c];
        fprintf(stdout, "music_eor_ch%i_tbl:\n", c + 1);
        for (vv = 0x81; vv < 0xff; ++vv) {
            n = note_ch_use_tbl[c][vv];
            if (n) {
                double vf0, vf1, vf, vfm, mf;
                mf = midi_freq[n & 0x7f];
                vf = vs / (255.0 - (double)vv);
                if (vf < mf) {
                    vf0 = vf;
                    vvo = vv + 1;
                    vf1 = vs / (255.0 - (double)(vvo));
                } else {
                    vf1 = vf;
                    vvo = vv - 1;
                    vf0 = vs / (255.0 - (double)(vvo));
                }
                vfm = (vf0 + vf1) / 2.0;
//printf("%i %02x %02x %f \t %f %c \t %f \t %f %c\n", c, vv, n, mf, vf0 - mf, (vf < mf) ? '*' : ' ', vfm - mf, vf1 - mf, (vf >= mf) ? '*' : ' ');
                if ((fabs(vfm - mf) < fabs(vf0 - mf)) && (fabs(vfm - mf) < fabs(vf1 - mf))) {
                    fprintf(stdout, "!by $%02x, $%02x  ;  %+.6f  %+.6f  %+.6f\n", vv, vvo ^ vv, vf0 - mf, vfm - mf, vf1 - mf);
                    bytes += 2;
                }
            }
        }
        fprintf(stdout, "!by $ff, $00\n");
        bytes += 2;
    }

    if (bytes > 0xff) {
        fprintf(stderr, "eor table too long!\n");
        return -1;
    }

    fprintf(stdout, "music_eor_tbl_tbl:\n"
                    "!by <(music_eor_ch1_tbl - music_eor_ch1_tbl)\n"
                    "!by <(music_eor_ch2_tbl - music_eor_ch1_tbl)\n"
                    "!by <(music_eor_ch3_tbl - music_eor_ch1_tbl)\n"
           );

    return bytes + 3;
}

/* ------------------------------------------------------------------------- */

static const char notes[12][3] = {
    "C-",
    "C#",
    "D-",
    "D#",
    "E-",
    "F-",
    "F#",
    "G-",
    "G#",
    "A-",
    "A#",
    "B-"
};

#define PATTERN_ROW_MAX 256
#define PATTERN_MAX 32
#define ORDER_MAX 256

typedef struct patternrow_s {
    BYTE delay;
    BYTE ch[3];
} patternrow_t;

typedef struct pattern_s {
    DWORD len;
    patternrow_t row[PATTERN_ROW_MAX];
} pattern_t;

#ifndef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#endif

#ifndef SORTB
#define SORTB(a,b) do { if ((a) > (b)) { BYTE t = (a); (a) = (b); (b) = t; }} while(0)
#endif

static int parse_midi_output(midi_t *midi, DWORD divider, DWORD patternlen, int tv)
{
    DWORD d;
    DWORD pnum = 0, ppos = 0, plen = 0;
    DWORD olen = 0;
    DWORD bytes;
    BYTE n;
    BYTE noteson[3] = { 0, 0, 0 };
    BYTE newon[3] = { 0, 0, 0 };
    BYTE notesoff[3] = { 0, 0, 0 };
    int c;
    midi_t *mrow;
    pattern_t pattern[PATTERN_MAX];
    BYTE order[ORDER_MAX];

    pattern[pnum].len = 0;

    memset(note_ch_use_tbl, 0, sizeof(note_ch_use_tbl));

    while (midi != NULL) {
        for (c = 0; c < 3; ++c) {
            newon[c] = noteson[c];
            notesoff[c] = 0;
        }

        /* scan noteoffs */
        d = 0;
        mrow = midi;
        while ((d == 0) && (mrow != NULL)) {
            n = mrow->note;
            if (!mrow->onoff) {
                /* find matching active note */
                for (c = 0; (newon[c] != n) && (c < 3); ++c);
                if (c == 3) {
                    fprintf(stderr, "could not find matching note for noteoff 0x%02x (%s%i)!\n", n, notes[n % 12], (n / 12));
                    return -1;
                } else {
                    newon[c] = 0;
                    notesoff[c] = n;
                }
            }
            mrow = mrow->next;
            if (mrow != NULL) {
                d = mrow->delay;
            }
        }

        /* scan noteons */
        d = 0;
        c = 0;
        mrow = midi;
        while ((d == 0) && (mrow != NULL)) {
            n = mrow->note;
            if (mrow->onoff == 1) {
                for (c = 0; (newon[c] != 0) && (c < 3); ++c);
                if (c == 3) {
                    fprintf(stderr, "could not find free channel for noteon 0x%02x (%s%i)!\n", n, notes[n % 12], (n / 12));
                    return -1;
                } else {
                    newon[c] = n;
                    if ((notesoff[c] == 1) && (noteson[c] == n)) {
                        /* allow reassignment */
                        noteson[c] = 0;
                    }
                    notesoff[c] = 0;
                }
            }
            mrow = mrow->next;
            if (mrow != NULL) {
                d = mrow->delay;
            }
        }

        /* sort noteons */
        if (noteson[0] && (noteson[0] == newon[0])) {
            SORTB(newon[1], newon[2]);
        } else if (noteson[1] && (noteson[1] == newon[1])) {
            SORTB(newon[0], newon[2]);
        } else if (noteson[2] && (noteson[2] == newon[2])) {
            SORTB(newon[0], newon[1]);
        } else {
            SORTB(newon[1], newon[2]);
            SORTB(newon[0], newon[1]);
            SORTB(newon[1], newon[2]);
        }

        /* move notes to lower channels for low notes */
        for (c = 1; c < 3; ++c) {
            int moved;
            moved = -1;
            n = newon[c];
            if (n
                && (n != noteson[c])
                && (midi_to_vic_tbl[c][n] == 1/*too low*/)) {
                if (newon[0] == 0) {
                    newon[0] = n;
                    newon[c] = 0;
                    moved = 0;
                } else if (newon[1] == 0) {
                    newon[1] = n;
                    newon[c] = 0;
                    moved = 1;
                } else {
                    fprintf(stdout, "; !!! %i %02x %s%i\n", c, n, notes[(n & 0x7f) % 12], ((n & 0x7f) / 12));
                }
            }
            /* resort if 2 moved to 0 and is higher */
            if ((c == 2) && (moved == 0) && (noteson[1] != newon[1])) {
                SORTB(newon[0], newon[1]);
            }
        }

        /* remember used notes */
        for (c = 0; c < 3; ++c) {
            note_ch_use_tbl[c][newon[c]] = 1;
        }

        /* store to pattern */
        for (c = 0; c < 3; ++c) {
            n = newon[c];
            if (n) {
                if (n == noteson[c]) {
                    pattern[pnum].row[plen].ch[c] = 0;
                } else {
                    pattern[pnum].row[plen].ch[c] = 0x80 | n;
                }
            } else {
                if (notesoff[c]) {
                    pattern[pnum].row[plen].ch[c] = notesoff[c];
                } else {
                    pattern[pnum].row[plen].ch[c] = 0;
                }
            }
        }

        if (d / divider >= 0x100) {
            fprintf(stderr, "too long delay %i!\n", d / divider);
            return -1;
        }

        pattern[pnum].row[plen].delay = d / divider;

#if 0
        fprintf(stdout, "%3i %3i %3i: ", pnum, ppos, plen);
        for (c = 0; c < 3; ++c) {
            n = pattern[pnum].row[plen].ch[c];
            if (n) {
                fprintf(stdout, "%s%i%c ", notes[(n & 0x7f) % 12], ((n & 0x7f) / 12), (n & 0x80) ? '+' : '-');
            } else {
                fprintf(stdout, "%s  ", "...");
            }
        }
        fprintf(stdout, " %i \n", pattern[pnum].row[plen].delay);

        {
            int dd = d / divider - 1;
            int i = 1;
            while (dd-- > 0) {
                fprintf(stdout, "%3i %3i %3i: ...  ...  ...\n", pnum, ppos + i, plen);
                ++i;
            }
        }
#endif

        for (c = 0; c < 3; ++c) {
            noteson[c] = newon[c];
        }

        if (++plen >= PATTERN_ROW_MAX) {
            fprintf(stderr, "too many rows in pattern!\n");
            return -1;
        }

        ppos += d / divider;
        if (ppos >= patternlen) {
            /* find matching pattern */
            DWORD pi;

            pattern[pnum].len = plen;

            for (pi = 0; pi < pnum; ++pi) {
                if ((pattern[pi].len == plen) && (memcmp(pattern[pi].row, pattern[pnum].row, plen * sizeof(patternrow_t)) == 0)) {
                    break;
                }
            }

            order[olen] = pi;

            if (++olen == ORDER_MAX) {
                fprintf(stderr, "too many orders!\n");
                return -1;
            }

            if (pi == pnum) {
                if (++pnum >= PATTERN_MAX) {
                    fprintf(stderr, "too many patterns!\n");
                    return -1;
                }
            }

            ppos = 0;
            plen = 0;
        }

        midi = mrow;
    }

    /* output last pattern (if partial) */
    if (ppos != 0) {
        DWORD pi;

        /* fill out missing time */
        if (ppos < patternlen) {
            pattern[pnum].row[plen - 1].delay += patternlen - ppos;
        }

        pattern[pnum].len = plen;

        /* find matching pattern */
        for (pi = 0; pi < pnum; ++pi) {
            if ((pattern[pi].len == plen) && (memcmp(pattern[pi].row, pattern[pnum].row, plen * sizeof(patternrow_t)) == 0)) {
                break;
            }
        }

        order[olen] = pi;

        if (pi == pnum) {
            ++pnum;
        }
    }

    /* compress and output patterns */
    {
        DWORD pi, o;
        bytes = 0;

        for (pi = 0; pi < pnum; ++pi) {
            fprintf(stdout, "music_pattern_%02x:\n", pi);
            plen = pattern[pi].len;
            for (ppos = 0; ppos < plen; ++ppos) {
                int odel;

                odel = pattern[pi].row[ppos].delay;

                fprintf(stdout, "!by $%02x",
                       odel % 0xf
                    | (pattern[pi].row[ppos].ch[0] ? 0x80 : 0)
                    | (pattern[pi].row[ppos].ch[1] ? 0x40 : 0)
                    | (pattern[pi].row[ppos].ch[2] ? 0x20 : 0)
                );
                ++bytes;
                for (c = 0; c < 3; ++c) {
                    n = pattern[pi].row[ppos].ch[c];
                    if (n) {
                        if (n & 0x80) {
                            BYTE vn = midi_to_vic_tbl[c][n & 0x7f];
                            fprintf(stdout, ", $%02x", vn);
                            /* remember used notes */
                            note_ch_use_tbl[c][vn] = n;
                        } else {
                            fprintf(stdout, ", $00");
                        }
                        ++bytes;
                    } else {
                        fprintf(stdout, "     ");
                    }
                }
                fprintf(stdout, "  ; ");
                for (c = 0; c < 3; ++c) {
                    n = pattern[pi].row[ppos].ch[c];
                    if (n) {
                        if (n & 0x80) {
                            fprintf(stdout, "%s%i%c ", notes[(n & 0x7f) % 12], ((n & 0x7f) / 12), (n & 0x80) ? '+' : '-');
                        } else {
                            fprintf(stdout, "---  ");
                        }
                    } else {
                        fprintf(stdout, "...  ");
                    }
                }
                fprintf(stdout, "\n");

                while (odel > 0xf) {
                    fprintf(stdout, "!by $0f                 ; delay\n");
                    ++bytes;
                    odel -= 0xf;
                }
            }
            fprintf(stdout, "!by 0\n");
            ++bytes;
        }

        bytes += 2 * pnum;
        fprintf(stdout, "music_pattern_addr_l:\n");
        for (pi = 0; pi < pnum; ++pi) {
            fprintf(stdout, "!by <music_pattern_%02x\n", pi);
        }
        fprintf(stdout, "music_pattern_addr_h:\n");
        for (pi = 0; pi < pnum; ++pi) {
            fprintf(stdout, "!by >music_pattern_%02x\n", pi);
        }

        bytes += olen;
        fprintf(stdout, "music_order:\n!by ");
        for (o = 0; o < olen; ++o) {
            if (o > 0) {
                fprintf(stdout, ", ");
            }
            fprintf(stdout, "$%02x", order[o]);
        }
        fprintf(stdout, ", $ff\n");
    }

    /* produce eor tables for used offtune notes */
    {
        int rc = output_vic_eor(tv);
        if (rc < 0) {
            return -1;
        }
        bytes += rc;
    }

    fprintf(stdout, "; %i b\n", bytes);
    return 0;
}

static int parse_midi(BYTE *midi_data, long midi_data_len, DWORD divider, DWORD patternlen, int transpose, int tv)
{
    long i = 0;
    int rc = 1;
    WORD ticks_per_beat;
    DWORD have_events[EVENT_TBL_SIZE];
    DWORD delays[DELAY_TBL_SIZE];
    midi_t *midi = NULL;

    if (midi_data_len < 24) {
        fprintf(stderr, "len %li too short\n", midi_data_len);
        goto fail;
    }

    if ((i = parse_midi_header(midi_data, midi_data_len, &ticks_per_beat)) < 0) {
        goto fail;
    }

    if ((midi = parse_midi_events(midi_data, midi_data_len, i, transpose)) == NULL) {
        goto fail;
    }

    if (parse_midi_scan_events(midi, have_events, delays, divider) < 0) {
        goto fail;
    }

    rc = parse_midi_output(midi, divider, patternlen, tv);

fail:
    free_midi_list(midi);
    midi = NULL;
    return rc;
}

/* ------------------------------------------------------------------------- */

static BYTE *midi_read(const char *filename, long *len_out)
{
    FILE *fp = NULL;
    BYTE *data = NULL;
    long len = 0;


    if ((fp = fopen(filename, "rb")) == NULL) {
        perror(filename);
        goto fail;
    }

    if (fseek(fp, 0, SEEK_END) != 0) {
        perror(filename);
        goto fail;
    }

    len = ftell(fp);

    fprintf(stderr, "%s: open, len %li\n", filename, len);

    rewind(fp);

    data = lib_malloc(len);

    if (fread(data, len, 1, fp) < 1) {
        fprintf(stderr, "%s: read error\n", filename);
        goto fail;
    }

    *len_out = len;

    fclose(fp);
    fp = NULL;

    return data;

fail:
    *len_out = 0;
    fclose(fp);
    fp = NULL;
    lib_free(data);
    data = NULL;
    return NULL;
}

/* ------------------------------------------------------------------------- */

static int cmdline_get_num(const char *str)
{
    int t;

    if (str == NULL) {
        return -1;
    }

    t = atoi(str);
    return t;
}

/* ------------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    DWORD divider = 1;
    DWORD patternlen = ~0;
    int transpose = 0;
    int isntsc = 0;
    BYTE *midi_data = NULL;
    long midi_data_len = 0;
    int rc = 0;


    if (argc < 2) {
        fprintf(stderr, "Usage: %s in.mid [DIV [PATTERNLEN [TRANSPOSE [ISNTSC]]]]\n", argv[0]);
        return 1;
    }

    if (argc >= 3) {
        if ((rc = cmdline_get_num(argv[2])) < 0) {
            return 2;
        }
        divider = (DWORD)rc;
        rc = 0;
    }

    if (argc >= 4) {
        if ((rc = cmdline_get_num(argv[3])) < 0) {
            return 3;
        }
        patternlen = (DWORD)rc;
        rc = 0;
    }

    if (argc >= 5) {
        transpose = atoi(argv[4]);
    }

    if (argc >= 6) {
        isntsc = atoi(argv[5]) ? 1 : 0;
    }

    if ((midi_data = midi_read(argv[1], &midi_data_len)) == NULL) {
        return 4;
    }

    midi_to_vic_precalc(isntsc);

    rc = parse_midi(midi_data, midi_data_len, divider, patternlen, transpose, isntsc);

    lib_free(midi_data);
    midi_data = NULL;

    return rc;
}
